/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * KVM/LOONGARCH: LOONGARCH specific KVM APIs
 *
 * Copyright (C) 2012-2014 Imagination Technologies Ltd.
 * Authors: Sanjay Lal <sanjayl@kymasys.com>
*/

#include "qemu/osdep.h"
#include <sys/ioctl.h>

#include <linux/kvm.h>

#include "qemu-common.h"
#include "cpu.h"
#include "internal.h"
#include "qemu/error-report.h"
#include "qemu/timer.h"
#include "qemu/main-loop.h"
#include "sysemu/sysemu.h"
#include "sysemu/kvm.h"
#include "sysemu/runstate.h"
#include "sysemu/cpus.h"
#include "kvm_larch.h"
#include "exec/memattrs.h"

#define DEBUG_KVM 0
/* A 16384-byte buffer can hold the 8-byte kvm_msrs header, plus
 * 2047 kvm_msr_entry structs */
#define MSR_BUF_SIZE 16384

#define DPRINTF(fmt, ...) \
    do { if (DEBUG_KVM) { fprintf(stderr, fmt, ## __VA_ARGS__); } } while (0)

/*
 * Define loongarch kvm version.
 * Add version number when
 * qemu/kvm interface changed
 */
#define KVM_LOONGARCH_VERSION KVM_LA_VERSION

static int kvm_loongarch_fpu_cap;
static int kvm_loongarch_lsx_cap;
static int kvm_loongarch_version_cap;

const KVMCapabilityInfo kvm_arch_required_capabilities[] = {
    KVM_CAP_LAST_INFO
};

static void kvm_loongarch_update_state(void *opaque, int running, RunState state);
static inline int kvm_larch_putq(CPUState *cs, uint64_t reg_id, uint64_t *addr);

unsigned long kvm_arch_vcpu_id(CPUState *cs)
{
    return cs->cpu_index;
}

int kvm_arch_init(MachineState *ms, KVMState *s)
{
    /* LOONGARCH has 128 signals */
    kvm_set_sigmask_len(s, 16);

    kvm_loongarch_fpu_cap = kvm_check_extension(s, KVM_CAP_LOONGARCH_FPU);
    kvm_loongarch_lsx_cap = kvm_check_extension(s, KVM_CAP_LOONGARCH_LSX);
    kvm_loongarch_version_cap = kvm_check_extension(s, KVM_CAP_LOONGARCH_VZ);

    if (kvm_loongarch_version_cap != KVM_LOONGARCH_VERSION) {
        warn_report("QEMU/KVM version not match, qemu_la_version: lvz-%d,\
                     kvm_la_version: lvz-%d \n",
                     KVM_LOONGARCH_VERSION, kvm_loongarch_version_cap);
    }
    DPRINTF("%s kvm_loongarch_fpu_cap:%d kvm_loongarch_lsx_cap:%d\n",
            __func__, kvm_loongarch_fpu_cap, kvm_loongarch_lsx_cap);
    return 0;
}

int kvm_arch_irqchip_create(KVMState *s)
{
    return 0;
}

int kvm_arch_init_vcpu(CPUState *cs)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int ret = 0;

    cpu->cpuStateEntry = qemu_add_vm_change_state_handler(kvm_loongarch_update_state, cs);

    if (kvm_loongarch_fpu_cap && env->CSR_MCSR1 & (1 << MCSR1_FP_SHIFT)) {
        ret = kvm_vcpu_enable_cap(cs, KVM_CAP_LOONGARCH_FPU, 0, 0);
        if (ret < 0) {
            /* mark unsupported so it gets disabled on reset */
            kvm_loongarch_fpu_cap = 0;
            ret = 0;
        }
    }

    if (kvm_loongarch_lsx_cap && env->CSR_MCSR1 & (1 << MCSR1_LSX_SHIFT)) {
        ret = kvm_vcpu_enable_cap(cs, KVM_CAP_LOONGARCH_LSX, 0, 0);
        if (ret < 0) {
            /* mark unsupported so it gets disabled on reset */
            kvm_loongarch_lsx_cap = 0;
            ret = 0;
        }
    }

    cpu->kvm_msr_buf = g_malloc0(MSR_BUF_SIZE);
    DPRINTF("%s\n", __func__);
    return ret;
}

int kvm_arch_destroy_vcpu(CPUState *cs)
{
    return 0;
}

static void kvm_msr_buf_reset(LOONGARCHCPU *cpu)
{
   memset(cpu->kvm_msr_buf, 0, MSR_BUF_SIZE);
}

static void kvm_msr_entry_add(LOONGARCHCPU *cpu, uint32_t index, uint64_t value)
{
    struct kvm_msrs *msrs = cpu->kvm_msr_buf;
    void *limit = ((void *)msrs) + MSR_BUF_SIZE;
    struct kvm_csr_entry *entry = &msrs->entries[msrs->ncsrs];

    assert((void *)(entry + 1) <= limit);

    entry->index = index;
    entry->reserved = 0;
    entry->data = value;
    msrs->ncsrs++;
}

void kvm_loongarch_reset_vcpu(LOONGARCHCPU *cpu)
{
    CPULOONGARCHState *env = &cpu->env;
    int ret = 0;
    uint64_t reset = 1;

    if (!kvm_loongarch_fpu_cap && env->CSR_MCSR1 & (1 << MCSR1_FP_SHIFT)) {
        warn_report("KVM does not support FPU, disabling");
        env->CSR_MCSR1 &= ~(1 << MCSR1_FP_SHIFT);
    }

    if (CPU(cpu)->kvm_fd > 0) {
        ret = kvm_larch_putq(CPU(cpu), KVM_REG_LOONGARCH_VCPU_RESET, &reset);
        if (ret < 0) {
            error_report("%s reset vcpu failed:%d", __func__, ret);
        }
    }

    DPRINTF("%s\n", __func__);
}

void kvm_arch_update_guest_debug(CPUState *cpu, struct kvm_guest_debug *dbg)
{
#ifdef KVM_GUESTDBG_USE_SW_BP
    if (kvm_sw_breakpoints_active(cpu)) {
        dbg->control |= KVM_GUESTDBG_ENABLE | KVM_GUESTDBG_USE_SW_BP;
    }
#endif
}

static const unsigned int brk_insn = 0x002b8005;

int kvm_arch_insert_sw_breakpoint(CPUState *cs, struct kvm_sw_breakpoint *bp)
{
    DPRINTF("%s\n", __func__);
    if (cpu_memory_rw_debug(cs, bp->pc, (uint8_t *)&bp->saved_insn, 4, 0) ||
        cpu_memory_rw_debug(cs, bp->pc, (uint8_t *)&brk_insn, 4, 1)) {
        error_report("%s failed", __func__);
        return -EINVAL;
    }
    return 0;
}

int kvm_arch_remove_sw_breakpoint(CPUState *cs, struct kvm_sw_breakpoint *bp)
{
    static uint32_t brk;

    DPRINTF("%s\n", __func__);
    if (cpu_memory_rw_debug(cs, bp->pc, (uint8_t *)&brk, 4, 0) ||
        brk != brk_insn ||
        cpu_memory_rw_debug(cs, bp->pc, (uint8_t *)&bp->saved_insn, 4, 1)) {
        error_report("%s failed", __func__);
        return -EINVAL;
    }
    return 0;
}

int kvm_arch_insert_hw_breakpoint(target_ulong addr,
                                  target_ulong len, int type)
{
    DPRINTF("%s\n", __func__);
    return 0;
}

int kvm_arch_remove_hw_breakpoint(target_ulong addr,
                                  target_ulong len, int type)
{
    DPRINTF("%s\n", __func__);
    return 0;
}

void kvm_arch_remove_all_hw_breakpoints(void)
{
    DPRINTF("%s\n", __func__);
}

static inline int cpu_loongarch_io_interrupts_pending(LOONGARCHCPU *cpu)
{
    CPULOONGARCHState *env = &cpu->env;

    return env->CSR_ESTAT & (0x1 << 2);
}

void kvm_arch_pre_run(CPUState *cs, struct kvm_run *run)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    int r;
    struct kvm_loongarch_interrupt intr;

    qemu_mutex_lock_iothread();

    if ((cs->interrupt_request & CPU_INTERRUPT_HARD) &&
            cpu_loongarch_io_interrupts_pending(cpu)) {
        intr.cpu = -1;
        intr.irq = 2;
        r = kvm_vcpu_ioctl(cs, KVM_INTERRUPT, &intr);
        if (r < 0) {
            error_report("%s: cpu %d: failed to inject IRQ %x",
                         __func__, cs->cpu_index, intr.irq);
        }
    }

    qemu_mutex_unlock_iothread();
}

MemTxAttrs kvm_arch_post_run(CPUState *cs, struct kvm_run *run)
{
    return MEMTXATTRS_UNSPECIFIED;
}

int kvm_arch_process_async_events(CPUState *cs)
{
    return cs->halted;
}

static bool kvm_loongarch_handle_debug(CPUState *cs)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;

    kvm_cpu_synchronize_state(cs);

    if (cs->singlestep_enabled) {
        return true;
    }
    if (kvm_find_sw_breakpoint(cs, env->active_tc.PC)) {
        return true;
    }

    return false;
}

int kvm_arch_handle_exit(CPUState *cs, struct kvm_run *run)
{
    int ret;

    DPRINTF("%s\n", __func__);
    switch (run->exit_reason) {
    case KVM_EXIT_HYPERCALL:
        DPRINTF("handle LOONGARCH hypercall\n");
        ret = 0;
        run->hypercall.ret = ret;
        break;

    case KVM_EXIT_DEBUG:
        ret = 0;
        if (kvm_loongarch_handle_debug(cs)) {
            ret = EXCP_DEBUG;
        }
        break;
    default:
        error_report("%s: unknown exit reason %d",
                     __func__, run->exit_reason);
        ret = -1;
        break;
    }

    return ret;
}

bool kvm_arch_stop_on_emulation_error(CPUState *cs)
{
    DPRINTF("%s\n", __func__);
    return true;
}
/*
#if 0
int kvmloongarch_load_kernel(CPUState *env, void *ram_base)
{
   int ret;

   ret = kvm_vcpu_ioctl(env, KVM_LOAD_KERNEL, ram_base);

   return ret;
}
#endif
*/
void kvm_arch_init_irq_routing(KVMState *s)
{
}

int kvm_loongarch_set_interrupt(LOONGARCHCPU *cpu, int irq, int level)
{
    CPUState *cs = CPU(cpu);
    struct kvm_loongarch_interrupt intr;

    if (!kvm_enabled()) {
        return 0;
    }

    intr.cpu = -1;

    if (level) {
        intr.irq = irq;
    } else {
        intr.irq = -irq;
    }

    kvm_vcpu_ioctl(cs, KVM_INTERRUPT, &intr);

    return 0;
}

int kvm_loongarch_set_ipi_interrupt(LOONGARCHCPU *cpu, int irq, int level)
{
    CPUState *cs = current_cpu;
    CPUState *dest_cs = CPU(cpu);
    struct kvm_loongarch_interrupt intr;

    if (!kvm_enabled()) {
        return 0;
    }

    intr.cpu = dest_cs->cpu_index;

    if (level) {
        intr.irq = irq;
    } else {
        intr.irq = -irq;
    }

    DPRINTF("%s:  IRQ: %d\n", __func__, intr.irq);
    if (!current_cpu) {
        cs = dest_cs;
    }
    kvm_vcpu_ioctl(cs, KVM_INTERRUPT, &intr);

    return 0;
}

static inline int kvm_loongarch_put_one_reg(CPUState *cs, uint64_t reg_id,
                                       int32_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_SET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_put_one_ureg(CPUState *cs, uint64_t reg_id,
                                        uint32_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_SET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_put_one_ulreg(CPUState *cs, uint64_t reg_id,
                                         target_ulong *addr)
{
    uint64_t val64 = *addr;
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)&val64
    };

    return kvm_vcpu_ioctl(cs, KVM_SET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_put_one_reg64(CPUState *cs, int64_t reg_id,
                                         int64_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_SET_ONE_REG, &csrreg);
}

static inline int kvm_larch_putq(CPUState *cs, uint64_t reg_id,
                                          uint64_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_SET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_get_one_reg(CPUState *cs, uint64_t reg_id,
                                       int32_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_GET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_get_one_ureg(CPUState *cs, uint64_t reg_id,
                                        uint32_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_GET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_get_one_ulreg(CPUState *cs, uint64_t reg_id,
                                         target_ulong *addr)
{
    int ret;
    uint64_t val64 = 0;
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)&val64
    };

    ret = kvm_vcpu_ioctl(cs, KVM_GET_ONE_REG, &csrreg);
    if (ret >= 0) {
        *addr = val64;
    }
    return ret;
}

static inline int kvm_loongarch_get_one_reg64(CPUState *cs, int64_t reg_id,
                                         int64_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_GET_ONE_REG, &csrreg);
}

static inline int kvm_larch_getq(CPUState *cs, uint64_t reg_id,
                                          uint64_t *addr)
{
    struct kvm_one_reg csrreg = {
        .id = reg_id,
        .addr = (uintptr_t)addr
    };

    return kvm_vcpu_ioctl(cs, KVM_GET_ONE_REG, &csrreg);
}

static inline int kvm_loongarch_change_one_reg(CPUState *cs, uint64_t reg_id,
                                          int32_t *addr, int32_t mask)
{
    int err;
    int32_t tmp, change;

    err = kvm_loongarch_get_one_reg(cs, reg_id, &tmp);
    if (err < 0) {
        return err;
    }

    /* only change bits in mask */
    change = (*addr ^ tmp) & mask;
    if (!change) {
        return 0;
    }

    tmp = tmp ^ change;
    return kvm_loongarch_put_one_reg(cs, reg_id, &tmp);
}

static inline int kvm_loongarch_change_one_reg64(CPUState *cs, uint64_t reg_id,
                                          int64_t *addr, int64_t mask)
{
    int err;
    int64_t tmp, change;

    err = kvm_loongarch_get_one_reg64(cs, reg_id, &tmp);
    if (err < 0) {
        DPRINTF("%s: Failed to get CSR_CONFIG7 (%d)\n", __func__, err);
        return err;
    }

    /* only change bits in mask */
    change = (*addr ^ tmp) & mask;
    if (!change) {
        return 0;
    }

    tmp = tmp ^ change;
    return kvm_loongarch_put_one_reg64(cs, reg_id, &tmp);
}
/*
 * Handle the VM clock being started or stopped
 */
static void kvm_loongarch_update_state(void *opaque, int running, RunState state)
{
    CPUState *cs = opaque;
    int ret;
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);

    /*
     * If state is already dirty (synced to QEMU) then the KVM timer state is
     * already saved and can be restored when it is synced back to KVM.
     */
    if (!running) {
        ret = kvm_larch_getq(cs,
              KVM_REG_LOONGARCH_COUNTER, &cpu->counter_value);
        if (ret < 0) {
            printf("%s: Failed to get counter_value (%d)\n", __func__, ret);
        }

    } else {
        ret = kvm_larch_putq(cs, KVM_REG_LOONGARCH_COUNTER,
                                         &(LOONGARCH_CPU(cs))->counter_value);
        if (ret < 0) {
            printf("%s: Failed to put counter_value (%d)\n", __func__, ret);
        }
    }
}

static int kvm_loongarch_put_fpu_registers(CPUState *cs, int level)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int err, ret = 0;
    unsigned int i;
    struct kvm_fpu fpu;

    /* Only put FPU state if we're emulating a CPU with an FPU */
    if (env->CSR_MCSR1 & (1 << MCSR1_FP_SHIFT)) {
        fpu.fcsr = env->active_fpu.fcsr0;
        for (i = 0; i < 32; i++) {
            memcpy(&fpu.fpr[i], &env->active_fpu.fpr[i], sizeof(struct kvm_fpureg));
        }
        for (i = 0; i < 8; i++) {
            ((char *)&fpu.fcc)[i] = env->active_fpu.cf[i];
        }
        fpu.vcsr = env->active_fpu.vcsr16;

        err = kvm_vcpu_ioctl(cs, KVM_SET_FPU, &fpu);
        if (err < 0) {
            DPRINTF("%s: Failed to get FPU (%d)\n", __func__, err);
            ret = err;
        }
    }

    return ret;
}

static int kvm_loongarch_get_fpu_registers(CPUState *cs)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int err, ret = 0;
    unsigned int i;
    struct kvm_fpu fpu;

    /* Only get FPU state if we're emulating a CPU with an FPU */
    if (env->CSR_MCSR1 & (1 << MCSR1_FP_SHIFT)) {
        err = kvm_vcpu_ioctl(cs, KVM_GET_FPU, &fpu);
        if (err < 0) {
            DPRINTF("%s: Failed to get FPU (%d)\n", __func__, err);
            ret = err;
        } else {
            env->active_fpu.fcsr0 = fpu.fcsr;
            for (i = 0; i < 32; i++) {
                memcpy(&env->active_fpu.fpr[i], &fpu.fpr[i], sizeof(struct kvm_fpureg));
            }
            for (i = 0; i < 8; i++) {
                env->active_fpu.cf[i] = ((char *)&fpu.fcc)[i];
            }
            env->active_fpu.vcsr16 = fpu.vcsr;
        }
    }

    return ret;
}

#define KVM_PUT_ONE_UREG64(cs, regidx, addr) \
    ({ \
        int err; \
        uint64_t csrid = 0; \
        csrid = (KVM_IOC_CSRID(regidx)); \
        err = kvm_larch_putq(cs, csrid, addr); \
        if (err < 0) { \
            DPRINTF("%s: Failed to put regidx 0x%x err:%d\n", __func__, regidx, err); \
        } \
        err; \
    })


static int kvm_loongarch_put_csr_registers(CPUState *cs, int level)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int ret = 0;

    (void)level;

    kvm_msr_buf_reset(cpu);

    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CRMD, env->CSR_CRMD);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRMD, env->CSR_PRMD);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_EUEN, env->CSR_EUEN);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MISC, env->CSR_MISC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ECFG, env->CSR_ECFG);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ESTAT, env->CSR_ESTAT);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERA, env->CSR_ERA);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_BADV, env->CSR_BADV);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_BADI, env->CSR_BADI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_EEPN, env->CSR_EEPN);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBIDX, env->CSR_TLBIDX);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBEHI, env->CSR_TLBEHI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBELO0, env->CSR_TLBELO0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBELO1, env->CSR_TLBELO1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GTLBC, env->CSR_GTLBC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TRGP, env->CSR_TRGP);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ASID, env->CSR_ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGDL, env->CSR_PGDL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGDH, env->CSR_PGDH);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGD, env->CSR_PGD);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PWCTL0, env->CSR_PWCTL0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PWCTL1, env->CSR_PWCTL1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_STLBPGSIZE, env->CSR_STLBPGSIZE);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_RVACFG, env->CSR_RVACFG);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CPUID, env->CSR_CPUID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG1, env->CSR_PRCFG1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG2, env->CSR_PRCFG2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG3, env->CSR_PRCFG3);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS0, env->CSR_KS0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS1, env->CSR_KS1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS2, env->CSR_KS2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS3, env->CSR_KS3);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS4, env->CSR_KS4);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS5, env->CSR_KS5);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS6, env->CSR_KS6);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS7, env->CSR_KS7);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TMID, env->CSR_TMID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CNTC, env->CSR_CNTC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TINTCLR, env->CSR_TINTCLR);

    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GSTAT, env->CSR_GSTAT);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GCFG, env->CSR_GCFG);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GINTC, env->CSR_GINTC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GCNTC, env->CSR_GCNTC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_LLBCTL, env->CSR_LLBCTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IMPCTL1, env->CSR_IMPCTL1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IMPCTL2, env->CSR_IMPCTL2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GNMI, env->CSR_GNMI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRENT, env->CSR_TLBRENT);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRBADV, env->CSR_TLBRBADV);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRERA, env->CSR_TLBRERA);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRSAVE, env->CSR_TLBRSAVE);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRELO0, env->CSR_TLBRELO0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRELO1, env->CSR_TLBRELO1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBREHI, env->CSR_TLBREHI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRPRMD, env->CSR_TLBRPRMD);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRCTL, env->CSR_ERRCTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRINFO, env->CSR_ERRINFO);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRINFO1, env->CSR_ERRINFO1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRENT, env->CSR_ERRENT);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRERA, env->CSR_ERRERA);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRSAVE, env->CSR_ERRSAVE);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CTAG, env->CSR_CTAG);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR0, env->CSR_MCSR0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR1, env->CSR_MCSR1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR2, env->CSR_MCSR2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR3, env->CSR_MCSR3);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR8, env->CSR_MCSR8);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR9, env->CSR_MCSR9);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR10, env->CSR_MCSR10);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR24, env->CSR_MCSR24);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN, env->CSR_UCAWIN);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN0_LO, env->CSR_UCAWIN0_LO);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN0_HI, env->CSR_UCAWIN0_HI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN1_LO, env->CSR_UCAWIN1_LO);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN1_HI, env->CSR_UCAWIN1_HI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN2_LO, env->CSR_UCAWIN2_LO);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN2_HI, env->CSR_UCAWIN2_HI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN3_LO, env->CSR_UCAWIN3_LO);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN3_HI, env->CSR_UCAWIN3_HI);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN0, env->CSR_DMWIN0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN1, env->CSR_DMWIN1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN2, env->CSR_DMWIN2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN3, env->CSR_DMWIN3);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL0, env->CSR_PERFCTRL0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR0, env->CSR_PERFCNTR0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL1, env->CSR_PERFCTRL1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR1, env->CSR_PERFCNTR1);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL2, env->CSR_PERFCTRL2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR2, env->CSR_PERFCNTR2);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL3, env->CSR_PERFCTRL3);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR3, env->CSR_PERFCNTR3);

    /* debug */
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MWPC, env->CSR_MWPC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MWPS, env->CSR_MWPS);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0ADDR, env->CSR_DB0ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0MASK, env->CSR_DB0MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0CTL, env->CSR_DB0CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0ASID, env->CSR_DB0ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1ADDR, env->CSR_DB1ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1MASK, env->CSR_DB1MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1CTL, env->CSR_DB1CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1ASID, env->CSR_DB1ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2ADDR, env->CSR_DB2ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2MASK, env->CSR_DB2MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2CTL, env->CSR_DB2CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2ASID, env->CSR_DB2ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3ADDR, env->CSR_DB3ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3MASK, env->CSR_DB3MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3CTL, env->CSR_DB3CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3ASID, env->CSR_DB3ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_FWPC, env->CSR_FWPC);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_FWPS, env->CSR_FWPS);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0ADDR, env->CSR_IB0ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0MASK, env->CSR_IB0MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0CTL, env->CSR_IB0CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0ASID, env->CSR_IB0ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1ADDR, env->CSR_IB1ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1MASK, env->CSR_IB1MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1CTL, env->CSR_IB1CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1ASID, env->CSR_IB1ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2ADDR, env->CSR_IB2ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2MASK, env->CSR_IB2MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2CTL, env->CSR_IB2CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2ASID, env->CSR_IB2ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3ADDR, env->CSR_IB3ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3MASK, env->CSR_IB3MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3CTL, env->CSR_IB3CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3ASID, env->CSR_IB3ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4ADDR, env->CSR_IB4ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4MASK, env->CSR_IB4MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4CTL, env->CSR_IB4CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4ASID, env->CSR_IB4ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5ADDR, env->CSR_IB5ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5MASK, env->CSR_IB5MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5CTL, env->CSR_IB5CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5ASID, env->CSR_IB5ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6ADDR, env->CSR_IB6ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6MASK, env->CSR_IB6MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6CTL, env->CSR_IB6CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6ASID, env->CSR_IB6ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7ADDR, env->CSR_IB7ADDR);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7MASK, env->CSR_IB7MASK);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7CTL, env->CSR_IB7CTL);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7ASID, env->CSR_IB7ASID);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DEBUG, env->CSR_DEBUG);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DERA, env->CSR_DERA);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DESAVE, env->CSR_DESAVE);

    ret = kvm_vcpu_ioctl(cs, KVM_SET_MSRS, cpu->kvm_msr_buf);
    if (ret < cpu->kvm_msr_buf->ncsrs) {
        struct kvm_csr_entry *e = &cpu->kvm_msr_buf->entries[ret];
        printf("error: failed to set CSR 0x%" PRIx32 " to 0x%" PRIx64"\n",
                     (uint32_t)e->index, (uint64_t)e->data);
    }

    /*
     * timer cfg must be put at last since it is used to enable
     * guest timer
     */
    ret |= KVM_PUT_ONE_UREG64(cs, LOONGARCH_CSR_TVAL, &env->CSR_TVAL);
    ret |= KVM_PUT_ONE_UREG64(cs, LOONGARCH_CSR_TCFG, &env->CSR_TCFG);
    return ret;
}

#define KVM_GET_ONE_UREG64(cs, regidx, addr) \
    ({ \
        int err; \
        uint64_t csrid = 0; \
        csrid = (KVM_IOC_CSRID(regidx)); \
        err = kvm_larch_getq(cs, csrid, addr); \
        if (err < 0) { \
            DPRINTF("%s: Failed to put regidx 0x%x err:%d\n", __func__, regidx, err); \
        } \
       err; \
    })

static int kvm_loongarch_get_csr_registers(CPUState *cs)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int ret = 0, i;
    struct kvm_csr_entry *csrs = cpu->kvm_msr_buf->entries;

    kvm_msr_buf_reset(cpu);

    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CRMD, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRMD, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_EUEN, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MISC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ECFG, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ESTAT, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERA, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_BADV, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_BADI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_EEPN, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBIDX, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBEHI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBELO0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBELO1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GTLBC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TRGP, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGDL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGDH, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PGD, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PWCTL0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PWCTL1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_STLBPGSIZE, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_RVACFG, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CPUID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PRCFG3, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS3, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS4, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS5, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS6, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_KS7, 0);

    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TMID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CNTC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TINTCLR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GSTAT, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GCFG, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GINTC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GCNTC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_LLBCTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IMPCTL1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IMPCTL2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_GNMI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRENT, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRBADV, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRERA, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRSAVE, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRELO0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRELO1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBREHI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_TLBRPRMD, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRCTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRINFO, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRINFO1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRENT, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRERA, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_ERRSAVE, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_CTAG, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR3, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR8, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR9, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR10, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MCSR24, 0);

    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN0_LO, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN0_HI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN1_LO, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN1_HI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN2_LO, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN2_HI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN3_LO, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_UCAWIN3_HI, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DMWIN3, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR0, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR1, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR2, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCTRL3, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_PERFCNTR3, 0);

    /* debug */
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MWPC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_MWPS, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB0ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB1ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB2ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DB3ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_FWPC, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_FWPS, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB0ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB1ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB2ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB3ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB4ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB5ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB6ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7ADDR, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7MASK, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7CTL, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_IB7ASID, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DEBUG, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DERA, 0);
    kvm_msr_entry_add(cpu, LOONGARCH_CSR_DESAVE, 0);

    ret = kvm_vcpu_ioctl(cs, KVM_GET_MSRS, cpu->kvm_msr_buf);
    if (ret < cpu->kvm_msr_buf->ncsrs) {
        struct kvm_csr_entry *e = &cpu->kvm_msr_buf->entries[ret];
        printf("error: failed to get CSR 0x%" PRIx32"\n",
                     (uint32_t)e->index);
    }

    for (i = 0; i < ret; i++) {
        uint32_t index = csrs[i].index;

        switch (index) {
        case LOONGARCH_CSR_CRMD:
            env->CSR_CRMD = csrs[i].data;
            break;
        case LOONGARCH_CSR_PRMD:
            env->CSR_PRMD = csrs[i].data;
            break;
        case LOONGARCH_CSR_EUEN:
            env->CSR_EUEN = csrs[i].data;
            break;
        case LOONGARCH_CSR_MISC:
            env->CSR_MISC = csrs[i].data;
            break;
        case LOONGARCH_CSR_ECFG:
            env->CSR_ECFG = csrs[i].data;
            break;
        case LOONGARCH_CSR_ESTAT:
            env->CSR_ESTAT = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERA:
            env->CSR_ERA = csrs[i].data;
            break;
        case LOONGARCH_CSR_BADV:
            env->CSR_BADV = csrs[i].data;
            break;
        case LOONGARCH_CSR_BADI:
            env->CSR_BADI = csrs[i].data;
            break;
        case LOONGARCH_CSR_EEPN:
            env->CSR_EEPN = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBIDX:
            env->CSR_TLBIDX = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBEHI:
            env->CSR_TLBEHI = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBELO0:
            env->CSR_TLBELO0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBELO1:
            env->CSR_TLBELO1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_GTLBC:
            env->CSR_GTLBC = csrs[i].data;
            break;
        case LOONGARCH_CSR_TRGP:
            env->CSR_TRGP = csrs[i].data;
            break;
        case LOONGARCH_CSR_ASID:
            env->CSR_ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_PGDL:
            env->CSR_PGDL = csrs[i].data;
            break;
        case LOONGARCH_CSR_PGDH:
            env->CSR_PGDH = csrs[i].data;
            break;
        case LOONGARCH_CSR_PGD:
            env->CSR_PGD = csrs[i].data;
            break;
        case LOONGARCH_CSR_PWCTL0:
            env->CSR_PWCTL0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PWCTL1:
            env->CSR_PWCTL1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_STLBPGSIZE:
            env->CSR_STLBPGSIZE = csrs[i].data;
            break;
        case LOONGARCH_CSR_RVACFG:
            env->CSR_RVACFG = csrs[i].data;
            break;
        case LOONGARCH_CSR_CPUID:
            env->CSR_CPUID = csrs[i].data;
            break;
        case LOONGARCH_CSR_PRCFG1:
            env->CSR_PRCFG1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PRCFG2:
            env->CSR_PRCFG2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PRCFG3:
            env->CSR_PRCFG3 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS0:
            env->CSR_KS0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS1:
            env->CSR_KS1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS2:
            env->CSR_KS2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS3:
            env->CSR_KS3 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS4:
            env->CSR_KS4 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS5:
            env->CSR_KS5 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS6:
            env->CSR_KS6 = csrs[i].data;
            break;
        case LOONGARCH_CSR_KS7:
            env->CSR_KS7 = csrs[i].data;
            break;

        case LOONGARCH_CSR_TMID:
            env->CSR_TMID = csrs[i].data;
            break;
        case LOONGARCH_CSR_CNTC:
            env->CSR_CNTC = csrs[i].data;
            break;
        case LOONGARCH_CSR_TINTCLR:
            env->CSR_TINTCLR = csrs[i].data;
            break;
        case LOONGARCH_CSR_GSTAT:
            env->CSR_GSTAT = csrs[i].data;
            break;
        case LOONGARCH_CSR_GCFG:
            env->CSR_GCFG = csrs[i].data;
            break;
        case LOONGARCH_CSR_GINTC:
            env->CSR_GINTC = csrs[i].data;
            break;
        case LOONGARCH_CSR_GCNTC:
            env->CSR_GCNTC = csrs[i].data;
            break;
        case LOONGARCH_CSR_LLBCTL:
            env->CSR_LLBCTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IMPCTL1:
            env->CSR_IMPCTL1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_IMPCTL2:
            env->CSR_IMPCTL2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_GNMI:
            env->CSR_GNMI = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRENT:
            env->CSR_TLBRENT = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRBADV:
            env->CSR_TLBRBADV = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRERA:
            env->CSR_TLBRERA = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRSAVE:
            env->CSR_TLBRSAVE = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRELO0:
            env->CSR_TLBRELO0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRELO1:
            env->CSR_TLBRELO1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBREHI:
            env->CSR_TLBREHI = csrs[i].data;
            break;
        case LOONGARCH_CSR_TLBRPRMD:
            env->CSR_TLBRPRMD = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRCTL:
            env->CSR_ERRCTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRINFO:
            env->CSR_ERRINFO = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRINFO1:
            env->CSR_ERRINFO1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRENT:
            env->CSR_ERRENT = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRERA:
            env->CSR_ERRERA = csrs[i].data;
            break;
        case LOONGARCH_CSR_ERRSAVE:
            env->CSR_ERRSAVE = csrs[i].data;
            break;
        case LOONGARCH_CSR_CTAG:
            env->CSR_CTAG = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR0:
            env->CSR_MCSR0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR1:
            env->CSR_MCSR1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR2:
            env->CSR_MCSR2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR3:
            env->CSR_MCSR3 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR8:
            env->CSR_MCSR8 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR9:
            env->CSR_MCSR9 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR10:
            env->CSR_MCSR10 = csrs[i].data;
            break;
        case LOONGARCH_CSR_MCSR24:
            env->CSR_MCSR24 = csrs[i].data;
            break;

        case LOONGARCH_CSR_UCAWIN:
            env->CSR_UCAWIN = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN0_LO:
            env->CSR_UCAWIN0_LO = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN0_HI:
            env->CSR_UCAWIN0_HI = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN1_LO:
            env->CSR_UCAWIN1_LO = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN1_HI:
            env->CSR_UCAWIN1_HI = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN2_LO:
            env->CSR_UCAWIN2_LO = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN2_HI:
            env->CSR_UCAWIN2_HI = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN3_LO:
            env->CSR_UCAWIN3_LO = csrs[i].data;
            break;
        case LOONGARCH_CSR_UCAWIN3_HI:
            env->CSR_UCAWIN3_HI = csrs[i].data;
            break;
        case LOONGARCH_CSR_DMWIN0:
            env->CSR_DMWIN0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_DMWIN1:
            env->CSR_DMWIN1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_DMWIN2:
            env->CSR_DMWIN2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_DMWIN3:
            env->CSR_DMWIN3 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCTRL0:
            env->CSR_PERFCTRL0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCNTR0:
            env->CSR_PERFCNTR0 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCTRL1:
            env->CSR_PERFCTRL1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCNTR1:
            env->CSR_PERFCNTR1 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCTRL2:
            env->CSR_PERFCTRL2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCNTR2:
            env->CSR_PERFCNTR2 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCTRL3:
            env->CSR_PERFCTRL3 = csrs[i].data;
            break;
        case LOONGARCH_CSR_PERFCNTR3:
            env->CSR_PERFCNTR3 = csrs[i].data;
            break;

        case LOONGARCH_CSR_MWPC:
            env->CSR_MWPC = csrs[i].data;
            break;
        case LOONGARCH_CSR_MWPS:
            env->CSR_MWPS = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB0ADDR:
            env->CSR_DB0ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB0MASK:
            env->CSR_DB0MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB0CTL:
            env->CSR_DB0CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB0ASID:
            env->CSR_DB0ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB1ADDR:
            env->CSR_DB1ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB1MASK:
            env->CSR_DB1MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB1CTL:
            env->CSR_DB1CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB1ASID:
            env->CSR_DB1ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB2ADDR:
            env->CSR_DB2ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB2MASK:
            env->CSR_DB2MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB2CTL:
            env->CSR_DB2CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB2ASID:
            env->CSR_DB2ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB3ADDR:
            env->CSR_DB3ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB3MASK:
            env->CSR_DB3MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB3CTL:
            env->CSR_DB3CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_DB3ASID:
            env->CSR_DB3ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_FWPC:
            env->CSR_FWPC = csrs[i].data;
            break;
        case LOONGARCH_CSR_FWPS:
            env->CSR_FWPS = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB0ADDR:
            env->CSR_IB0ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB0MASK:
            env->CSR_IB0MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB0CTL:
            env->CSR_IB0CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB0ASID:
            env->CSR_IB0ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB1ADDR:
            env->CSR_IB1ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB1MASK:
            env->CSR_IB1MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB1CTL:
            env->CSR_IB1CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB1ASID:
            env->CSR_IB1ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB2ADDR:
            env->CSR_IB2ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB2MASK:
            env->CSR_IB2MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB2CTL:
            env->CSR_IB2CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB2ASID:
            env->CSR_IB2ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB3ADDR:
            env->CSR_IB3ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB3MASK:
            env->CSR_IB3MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB3CTL:
            env->CSR_IB3CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB3ASID:
            env->CSR_IB3ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB4ADDR:
            env->CSR_IB4ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB4MASK:
            env->CSR_IB4MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB4CTL:
            env->CSR_IB4CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB4ASID:
            env->CSR_IB4ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB5ADDR:
            env->CSR_IB5ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB5MASK:
            env->CSR_IB5MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB5CTL:
            env->CSR_IB5CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB5ASID:
            env->CSR_IB5ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB6ADDR:
            env->CSR_IB6ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB6MASK:
            env->CSR_IB6MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB6CTL:
            env->CSR_IB6CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB6ASID:
            env->CSR_IB6ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB7ADDR:
            env->CSR_IB7ADDR = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB7MASK:
            env->CSR_IB7MASK = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB7CTL:
            env->CSR_IB7CTL = csrs[i].data;
            break;
        case LOONGARCH_CSR_IB7ASID:
            env->CSR_IB7ASID = csrs[i].data;
            break;
        case LOONGARCH_CSR_DEBUG:
            env->CSR_DEBUG = csrs[i].data;
            break;
        case LOONGARCH_CSR_DERA:
            env->CSR_DERA = csrs[i].data;
            break;
        case LOONGARCH_CSR_DESAVE:
            env->CSR_DESAVE = csrs[i].data;
            break;
        default:
            break;
        }
    }

    ret |= KVM_GET_ONE_UREG64(cs, LOONGARCH_CSR_TVAL, &env->CSR_TVAL);
    ret |= KVM_GET_ONE_UREG64(cs, LOONGARCH_CSR_TCFG, &env->CSR_TCFG);
    return ret;
}

int kvm_arch_put_registers(CPUState *cs, int level)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    struct kvm_regs regs;
    int ret;
    int i;

   /* Set the registers based on QEMU's view of things */
    for (i = 0; i < 32; i++) {
        regs.gpr[i] = (int64_t)(target_long)env->active_tc.gpr[i];
    }

    regs.pc = (int64_t)(target_long)env->active_tc.PC;

    ret = kvm_vcpu_ioctl(cs, KVM_SET_REGS, &regs);

    if (ret < 0) {
        return ret;
    }

    ret = kvm_loongarch_put_csr_registers(cs, level);
    if (ret < 0) {
        return ret;
    }

    ret = kvm_loongarch_put_fpu_registers(cs, level);
    if (ret < 0) {
        return ret;
    }

    return ret;
}

int kvm_arch_get_registers(CPUState *cs)
{
    LOONGARCHCPU *cpu = LOONGARCH_CPU(cs);
    CPULOONGARCHState *env = &cpu->env;
    int ret = 0;
    struct kvm_regs regs;
    int i;

    /* Get the current register set as KVM seems it */
    ret = kvm_vcpu_ioctl(cs, KVM_GET_REGS, &regs);

    if (ret < 0) {
        return ret;
    }

    for (i = 0; i < 32; i++) {
        env->active_tc.gpr[i] = regs.gpr[i];
    }

    env->active_tc.PC = regs.pc;

    kvm_loongarch_get_csr_registers(cs);
    kvm_loongarch_get_fpu_registers(cs);

    return ret;
}

int kvm_arch_fixup_msi_route(struct kvm_irq_routing_entry *route,
                             uint64_t address, uint32_t data, PCIDevice *dev)
{
    return 0;
}

int kvm_arch_add_msi_route_post(struct kvm_irq_routing_entry *route,
                                int vector, PCIDevice *dev)
{
    return 0;
}

int kvm_arch_release_virq_post(int virq)
{
    return 0;
}

int kvm_arch_msi_data_to_gsi(uint32_t data)
{
    abort();
}
